/*-------------------------------------------------------------------------
 Hash function Luffa

 File name   : Luffa_Micro.v
 Version     : Version 0.2
 Created     : FEB/08/2010
 Last update : FEB/14/2010
 Desgined by : Akashi Satohr
 
 
 Copyright (C) 2010 AIST
 
 By using this code, you agree to the following terms and conditions.
 
 This code is copyrighted by AIST ("us").
 
 Permission is hereby granted to copy, reproduce, redistribute or
 otherwise use this code as long as: there is no monetary profit gained
 specifically from the use or reproduction of this code, it is not sold,
 rented, traded or otherwise marketed, and this copyright notice is
 included prominently in any copy made.
 
 We shall not be liable for any damages, including without limitation
 direct, indirect, incidental, special or consequential damages arising
 from the use of this code.
 
 When you publish any results arising from the use of this code, we will
 appreciate it if you can cite our webpage
 (http://www.rcis.aist.go.jp/special/SASEBO/index-en.html).
 -------------------------------------------------------------------------*/

//`timescale 1ns / 1ps
//`define TBL


//////////////////////////
//      Main Module     //
//////////////////////////

module Luffa(CLK, RSTn, EN, Mrdy, Init, EoM, M, BSY, H);
  input  CLK;                                               // System clock
  input  RSTn;                                              // Negative reset
  input  EN;                                                // Enable
  input  Mrdy;                                              // Set High for message input
  input  Init;                                              // Set High for the first message input
  input  EoM;                                               // Set High for the last message input
  input  [255:0] M;                                         // Message input
  output BSY;                                               // Busy flag
  output [255:0] H;                                         // Hash output

  parameter V0 = 256'h6d251e69_44b051e0_4eaa6fb4_dbf78465_6e292011_90152df4_ee058139_def610bb;
  parameter V1 = 256'hc3b44b95_d9d2f256_70eee9a0_de099fa3_5d9b0557_8fc944b3_cf1ccf0e_746cd581;
  parameter V2 = 256'hf7efc89d_5dba5781_04016ce5_ad659c05_0306194f_666d1836_24aa230a_8b264ae7;

  wire [255:0] Min;                                         // Message input for MI function
  wire [255:0] MI   [0:2];                                  // Message Injection function input
  wire [255:0] MO   [0:2];                                  // Message Injection function output
  reg  [255:0] Qrg  [0:2];                                  // Work registers
  reg  [31:0]  CLrg [0:2];                                  // Constatn registers
  reg  [31:0]  CRrg [0:2];                                  // Constatn registers
  wire [31:0]  CL,  CR;                                     // Current constants
  wire [31:0]  CLx, CRx;                                    // Next constants
  reg  BSYrg,  FINrg;                                       // Busy, Finalization registers
  reg  [8:0]   RNDrg;                                       // Round register for Step function
  reg  [3:0]   PIPrg;                                       // Pipeline control register
  wire [1:0]   PLrnd;                                       // = PIPrg[1:0]
  wire [1:0]   SCrnd;                                       // SubCrumb round
  wire [1:0]   MWrnd;                                       // MixWord round 

  wire [15:0] Qi [15:0];                                    // SubScrumb input
  wire [15:0] Si [3:0];                                     // SubScrumb input
  wire [15:0] So [3:0];                                     // SubScrumb output
  wire [31:0] Ri [7:0];                                     // MixWord input
  wire [31:0] Xi [1:0];                                     // MixWord input
  wire [31:0] Xo [1:0];                                     // MixWord output
  wire [31:0] P  [1:0];                                     // Constant XORed
  wire Xfg;                                                 // Constant XOR flag
  wire [31:0] C0, CP;

  assign Min   = (FINrg)? 256'h0: M;
  assign MI[0] = (Init)? V0: Qrg[0];
  assign MI[1] = (Init)? V1: Qrg[1];
  assign MI[2] = (Init)? V2: Qrg[2];

  MI MX(Min, MI[0], MI[1], MI[2], H, MO[0], MO[1], MO[2]);

  assign SCrnd = PIPrg[3:2];
  assign MWrnd = (PIPrg[3:2] - 1) % 4;
  assign PLrnd = PIPrg[1:0];
  assign {Qi[15], Qi[14], Qi[13], Qi[12], Qi[11], Qi[10], Qi[9], Qi[8], Qi[7], Qi[6], Qi[5], Qi[4], Qi[3], Qi[2], Qi[1], Qi[0]}
         = (SCrnd < 3)? MI[SCrnd]: MI[2];
  assign {Si[3], Si[2], Si[1], Si[0]} = (PLrnd == 2'd0)? {Qi[15], Qi[13], Qi[11], Qi[9]}:
                                        (PLrnd == 2'd1)? {Qi[14], Qi[12], Qi[10], Qi[8]}:
                                        (PLrnd == 2'd2)? {Qi[ 5], Qi[ 3], Qi[ 1], Qi[7]}:
                                                         {Qi[ 4], Qi[ 2], Qi[ 0], Qi[6]};
  `ifdef TBL
    SubCrumb_TBL SCX(Si[3], Si[2], Si[1], Si[0], So[3], So[2], So[1], So[0]);
  `else
    SubCrumb_BS  SCX(Si[3], Si[2], Si[1], Si[0], So[3], So[2], So[1], So[0]);
  `endif

  assign {Ri[7], Ri[6], Ri[5], Ri[4], Ri[3], Ri[2], Ri[1], Ri[0]} = (MWrnd < 3)? MI[MWrnd]: MI[2];
  assign {Xi[1], Xi[0]} = (PLrnd == 2'd0)? {Ri[7], Ri[3]}:
                          (PLrnd == 2'd1)? {Ri[6], Ri[2]}:
                          (PLrnd == 2'd2)? {Ri[5], Ri[1]}:
                                           {Ri[4], Ri[0]}; 
  MixWord MWX(Xi[1], Xi[0], Xo[1], Xo[0]);
    
  assign CL  = (MWrnd < 3)? CLrg[MWrnd]: CLrg[2];
  assign CR  = (MWrnd < 3)? CRrg[MWrnd]: CRrg[2];
  assign {C0, CP}   = {CL[30:0], CR[31:0], 1'b0} ^ ({64{CL[31]}} & 64'hc4d6496c55c61c8d);
  assign {CRx, CLx} = {CP[30:0], C0[31:0], 1'b0} ^ ({64{CP[31]}} & 64'hc4d6496c55c61c8d);

  assign Xfg = (PIPrg[3] | PIPrg[2]) & ~(PIPrg[1] | PIPrg[0]);
  assign {P[1], P[0]} = (Xfg == 0)? {Xo[1], Xo[0]}: {Xo[1], Xo[0]} ^ {C0, CRx};
  assign H = MI[0] ^ MI[1] ^ MI[2];
  assign BSY = ~RNDrg[0] | FINrg;

  always @(posedge CLK) begin
    if (RSTn == 0) begin
      BSYrg <= 0;
      FINrg <= 0;
      RNDrg <= 9'b000000001;
      PIPrg <= 4'd0;
    end
    else if (EN == 1) begin
      if (RNDrg[0] == 1) begin
        if (Mrdy | FINrg == 1) begin
          Qrg[0]  <= MO[0];
          Qrg[1]  <= MO[1];
          Qrg[2]  <= MO[2];
          CLrg[0] <= 32'h181cca53;
          CRrg[0] <= 32'h380cde06;
          CLrg[1] <= 32'h5b6f0876;
          CRrg[1] <= 32'hf16f8594;
          CLrg[2] <= 32'h7e106ce9;
          CRrg[2] <= 32'h38979cb0;
          RNDrg   <= {RNDrg[7:0], RNDrg[8]};
          FINrg   <= (FINrg)? 0: EoM;
        end
      end
      else begin
        if (Xfg == 1) begin
          CLrg[MWrnd] <= CLx;
          CRrg[MWrnd] <= CRx;
        end

        if (SCrnd != 3) begin
          case (PLrnd)
            2'b00:   {Qrg[SCrnd][255:240], Qrg[SCrnd][223:208], Qrg[SCrnd][191:176], Qrg[SCrnd][159:144]}
                       <= {So[3], So[2], So[1], So[0]};
            2'b01:   {Qrg[SCrnd][239:224], Qrg[SCrnd][207:192], Qrg[SCrnd][175:160], Qrg[SCrnd][143:128]}
                       <= {So[3], So[2], So[1], So[0]};
            2'b10:   {Qrg[SCrnd][ 95: 80], Qrg[SCrnd][ 63: 48], Qrg[SCrnd][ 31: 16], Qrg[SCrnd][127:112]}
                       <= {So[3], So[2], So[1], So[0]};
            default: {Qrg[SCrnd][ 79: 64], Qrg[SCrnd][ 47: 32], Qrg[SCrnd][ 15:  0], Qrg[SCrnd][111: 96]}
                       <= {So[3], So[2], So[1], So[0]};
          endcase
        end

        if (MWrnd != 3) begin
          case (PLrnd)
            2'b00:   {Qrg[MWrnd][255:224], Qrg[MWrnd][127: 96]} <= {P[1], P[0]};
            2'b01:   {Qrg[MWrnd][223:192], Qrg[MWrnd][ 95: 64]} <= {P[1], P[0]};
            2'b10:   {Qrg[MWrnd][191:160], Qrg[MWrnd][ 63: 32]} <= {P[1], P[0]};            
            default: {Qrg[MWrnd][159:128], Qrg[MWrnd][ 31:  0]} <= {P[1], P[0]};
          endcase
        end        

        if (PIPrg == 4'hf)
          RNDrg <= {RNDrg[7:0], RNDrg[8]};

        PIPrg <= PIPrg + 1; 
      end
    end
  end
endmodule


//////////////////////////
//   Message Injection  //
//////////////////////////
module MI(M, H0, H1, H2, H123, X0, X1, X2);
  input  [255:0] M, H0, H1, H2, H123;
  output [255:0] X0, X1, X2;

  wire [255:0] t, M2, M4, X1t, X2t;

  function [255:0] MUL2;
    input [255:0] x;
      MUL2 = {x[31:0], x[255:224]^x[31:0], x[223:192], x[191:160]^x[31:0], x[159:128]^x[31:0], x[127:32]};
  endfunction

  function [255:0] TWEAK1;
    input [255:0] x;
      TWEAK1 = {x[255:128], {x[126:96], x[127]}, {x[94:64], x[95]}, {x[62:32],x[63]}, {x[30:0], x[31]}};
  endfunction

  function [255:0] TWEAK2;
    input [255:0] x;
      TWEAK2 = {x[255:128], {x[125:96], x[127:126]}, {x[93:64], x[95:94]}, {x[61:32],x[63:62]}, {x[29:0], x[31:30]}};
  endfunction

  assign t   = MUL2(H123);
  assign M2  = MUL2(M);
  assign M4  = MUL2(M2);
  assign X0  = H0 ^ M  ^ t;
  assign X1t = H1 ^ M2 ^ t;
  assign X2t = H2 ^ M4 ^ t;
  assign X1  = TWEAK1(X1t);
  assign X2  = TWEAK2(X2t);
endmodule


//////////////////////////
//  SubScrumb Bit Slice //
//////////////////////////
module SubCrumb_BS(a0, a1, a2, a3, x0, x1, x2, x3);
  input  [15:0] a0, a1, a2, a3;
  output [15:0] x0, x1, x2, x3;

  wire [15:0] t0, t1, t2, t3, t4, t5, t6, t7;

  assign t0 =  a0 | a1;
  assign t1 =  a2 ^ a3;
  assign t2 =  t0 ^ a3;
  assign t3 =  a0 & a3;
  assign t4 = ~a1 ^ t3;
  assign t5 =  t1 ^ t3;
  assign t6 =  t4 ^ (t1 & t2);
  assign t7 =  t4 | t5;
  assign x0 =  a0 ^ t7;
  assign x1 = ~t2 ^ t7;
  assign x2 =  t6 & t7;
  assign x3 =  t6 ^ t5;
endmodule


//////////////////////////
//   SubScrumb Table    //
//////////////////////////
module SubCrumb_TBL(a0, a1, a2, a3, x0, x1, x2, x3);
  input  [15:0] a0, a1, a2, a3;
  output [15:0] x0, x1, x2, x3;

  function [3:0] S;
  input    [3:0] a;
    case (a)
        0:S=13; 1:S=14; 2:S=0;  3:S=1;   4:S=5;   5:S=10; 6:S=7;  7:S=6;
        8:S=11; 9:S=3; 10:S=9; 11:S=12; 12:S=15; 13:S=8; 14:S=2; 15:S=4;
     endcase
  endfunction

  generate
    genvar i;
    for (i = 0; i < 16; i = i+1) begin
      assign {x3[i], x2[i], x1[i], x0[i]} = S({a3[i], a2[i], a1[i], a0[i]});
    end
  endgenerate
endmodule


//////////////////////////
//        MixWord       //
//////////////////////////
module MixWord(x0, x4, y0, y4);
  input  [31:0] x0, x4;
  output [31:0] y0, y4;

  wire [31:0] t0, t1, t2;

  assign t0 = x0 ^ x4;
  assign t1 = {x0[29:0], x0[31:30]} ^ t0;
  assign t2 = {t0[17:0], t0[31:18]} ^ t1;
  assign y0 = {t1[21:0], t1[31:22]} ^ t2;
  assign y4 = {t2[30:0], t2[31]};
endmodule